/* 
   a27d-x.c  2017.10.03 new gauge 3.

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_odeiv.h>
#define D_LIMIT_LOG 690.0   // log(10^300)  out of double
/*
#define RAPID_CHANGE (1+1e-10)    // y[0] changes rapidly.  >1
#define RAPID_CHANGE2 (1-1e-10)    // y[0] changes rapidly. <1
*/
#define RAPID_CHANGE (2)    // y[0] changes rapidly.  >1
#define RAPID_CHANGE2 (1/2)    // y[0] changes rapidly. <1

gsl_odeiv_control *setparam(int argc,char *argv[],double *tp,double *t1p,double *hp,double y[],int *show_diffp);
int output_value(double t,double y[],double t0,double h);
int output_value2(double t,double y[],double t0,double h);
int showdiff(int x,double t,double y[]);
int show_diff2_case1(int x, double t, double y[],double y_ans[]);
int output_im_value(double t,double y[]);

// Parameters k,n of H^k_n(t^2,z^2)
double Hkn_k=3;
double Hkn_n=1;
double Hkn_z=0;  // will be set to sqrt(10^8) as default.  z^2=y,  y=10^8;
//double Hkn_abs_err=1e-200;
double Hkn_abs_err=1e-10;
double Hkn_rel_err=1e-20;
double Hkn_c0=1.0;  /* scalar Gauge */
double Hkn_c1=1.0;
double Hkn_c2=1.0;
double Hkn_logc0=0.0;
double Hkn_logC=0.0;  /* log of global scalar Gauge factor */
// int Hkn_auto_abs_err=1;  // 2017.10.04 は使った.
int Hkn_auto_abs_err=0;
char *Hkn_fname="stdout";
int Verbose=1;

int Hkn_switch_gauge=1;  // 1: exp(-t^2+2zt+2*(z+1)*log(t)),  2: exp(Hkn_logc0)
char *Hkn_file=NULL;
int Hkn_file_from=0;     // 保存を開始する x の値.
int Hkn_file_max=10000;  // file へ保存するデータの最大量.
int Hkn_file_step=100;   // from, from+step,from+2*step, ...
int Hkn_i=0;
int Hkn_j=0; // pos in Hmat
int Hkn_init_gauge=0;

double max_abs(double y[]) {
  double maxy;
  double absy;
  int i;
  maxy = (y[0]>0?y[0]:-y[0]);
  for (i=1 ; i<3; i++) {
    absy = (y[i]>0?y[i]:-y[i]);
    if (absy > maxy) maxy = absy;
  }
  return(maxy);
}
// it is not used.
double cpower(double x,double k) {
  int kk,i;
  double val;
  kk = k;
  if (kk == 0) return(1.0);
  if (kk < 0) return(1/cpower(x,-k));
  val = 1.0;
  for (i=0; i<k; i++) val = val*x;
  return(val);
}
int func (double t, const double y[], double f[], void *params)
{
  double y_0,y_1,y_2;
  double n,k,z;
  double exparg;
  n = Hkn_n; k = Hkn_k; z = Hkn_z;
  y_0 = y[0]; y_1 = y[1]; y_2 = y[2]; 
  switch (Hkn_switch_gauge) {
    case 1:
      f[0]= +2*t*y_0-2*z*y_0-(2*(k+1)/t)*y_0
        +(2*exp(-Hkn_logc0)*y_1*Hkn_c1)/(Hkn_c0*t);
      break;
    case 2:
      //  f[0]= (2*exp(-t*t+2*t*z + 2*(k+1)*log(t))*y_1*(Hkn_c1/Hkn_c0))/t;
      exparg = -t*t +2*t*z-Hkn_logc0; // it will be <>0 when switched.
      exparg = exparg+2*(k+1)*log(t);
      if (exparg < -D_LIMIT_LOG) f[0] = 0.0;
      else f[0]= 2*exp(exparg)*y_1*Hkn_c1/(Hkn_c0*t);
      // printf("exparg=%.15e \n",exparg);
      break;
    default:
      fprintf(stderr,"Error: unknown Hkn_switch_gauge mode.\n");
      break;
  }
  f[1]= (y_2*(Hkn_c2/Hkn_c1))/t;
  f[2]= (-2*t*z*(2*n-1)*y_1*(Hkn_c1/Hkn_c2) 
         -(4*t*z+2*(n-1))*y_2)/t;
  return GSL_SUCCESS;
}

int main(int argc,char *argv[])
{
  /* アルゴリズム T を用いて, 次元 2 の方程式を解くステップ関数を生成 */
  /* Runge-Kutta-Fehlberg(4,5) */
  const gsl_odeiv_step_type *T = gsl_odeiv_step_rkf45;
  gsl_odeiv_step *s = gsl_odeiv_step_alloc(T, 3);

  /* 
   * 各ステップにおいて解の値 y_i(t) に対する誤差を絶対誤差 1e-20
   * 相対誤差 1e-10 の範囲内に抑えるようなステップ調整法を生成
   */
  //    gsl_odeiv_control *c = gsl_odeiv_control_y_new(1e-20, 1e-10);
  gsl_odeiv_control *c = gsl_odeiv_control_y_new(Hkn_abs_err, Hkn_rel_err);

  /* 4 次元の方程式のための求解関数の生成 */
  gsl_odeiv_evolve *e = gsl_odeiv_evolve_alloc(3);
  gsl_odeiv_system sys = {func, NULL, 3, NULL};
  /* t : start, t1 : goal */
  double t; 
  double t0;
  double t1;
  double h = 1e-6; /* ステップ幅の初期値 */

  // Default values.        
  t=2000; 
  Hkn_z = 10000; // y^2=z  10^5
  //  t1 = 4000;
  t1=2050;
/*  a30-test.m , a26-ppp.m  a30-test4-out.txt
test4initCase1[2000]
Out[8]//FortranForm= 
         List(1.0092527990727128287e15,0.000063078313247623732261,
     -   -0.0000315391568209)

  */
  double y[3]={1.0092527990727128287e15,0.000063078313247623732261,
                -0.0000315391568209};

  int show_diff=1;
  double tt;
  double case1_g_log=0.0;
  static int first=1;
  int x,xx,xx_save,i;
  double t_maxarg,k;
  double max_abs_y;
  static int auto_abs_err5=0;

  // default 値と異なる設定をする場合.
  if (argc>1) {
    show_diff=0;
    c=setparam(argc,argv,&t,&t1,&h,y,&show_diff);
    output_value(t,y,t,h);
    output_value2(t,y,t,h);
    if (c == NULL) return(0);
  }
  k = Hkn_k;
  Hkn_c0=y[0];Hkn_c1=y[1]; Hkn_c2=y[2];  
  y[0]=1.0; y[1]=1.0; y[2]=1.0;
  t0 = t;
        
  x = (int) t; xx = (int) t*t;

  while (t < t1) {
    int status = gsl_odeiv_evolve_apply(e, c, s, &sys, &t, t1, &h, y);
    if (status != GSL_SUCCESS) break;

    if (Hkn_auto_abs_err && (!auto_abs_err5)) {
      if ((max_abs_y=max_abs(y)) > (1e+15)*Hkn_abs_err) {
        Hkn_abs_err = max_abs_y*(1e-10);
        if (Hkn_abs_err > 1e-5) {
          Hkn_abs_err=1e-5;
          auto_abs_err5=1;
        }
        c = gsl_odeiv_control_y_new(Hkn_abs_err, Hkn_rel_err);
        if (Verbose) {
          printf("max_abs_y=%.15e, abs_err is changed to %.15e\n",max_abs_y,Hkn_abs_err);
        }
      }
    }

    if (t>x) {  /* Gauge を変更するか判断する */
      t_maxarg = (Hkn_z/2.0)*(1+sqrt(1+4*(k+1)/(Hkn_z*Hkn_z)));  /* tt が最大となるあたり. */
      tt=-t*t+2*t*Hkn_z+2*(Hkn_k+1)*log(t);
      if ((Hkn_switch_gauge==1) && (t > t_maxarg-1.0/1000.0)) {
        if (Verbose) printf("For t>t_maxarg=%.15e, -t^2+2tz+2(k+1)*log(t)=%.15e, switch to Hkn_switch_gauge=2\n",t_maxarg,tt);
        Hkn_switch_gauge=2;
        Hkn_logc0 = tt;
        // Gauge をここから const へ移行。
      }
      if (Hkn_switch_gauge==1) {case1_g_log=-t*t+2*t*Hkn_z+2*(k+1)*log(t);}
      else case1_g_log=0.0;
      if (Verbose && (show_diff < 2)){
        printf("t=%.15e t^2=%.15e, log(y[0]*gauge)=%.15e, y[1]=%.15e, y[2]=%.15e,", t,t*t, log(y[0]*Hkn_c0)+case1_g_log+Hkn_logc0, y[1]*Hkn_c1,y[2]*Hkn_c2);
        printf(" log(y[0]*Hkn_c0)=%.15e\n",log(y[0]*Hkn_c0));
      }
      if (isnan(y[0])) {
        printf("bare y[0],y[1],y[2]= %.15e, %.15e, %.15e\n",y[0],y[1],y[2]);
        printf("Hkn_c0=%.15e, case1_g_log=%.15e, Hkn_logc0=%.15e\n",Hkn_c0,case1_g_log,Hkn_logc0);
        printf("Error: break by nan\n");
        break;
      }
      if (show_diff) showdiff(x,t,y);
      x++;
    }
    if ((t*t) > xx) {
      if ((Hkn_file != NULL) && (xx > Hkn_file_from)) {
        if (first) {xx = (int) (t*t); first=0;}
        xx_save=xx;
        for (i=0; i< ((int)(t*t)-xx_save)+1; i++) {output_im_value(t,y);xx++;}
      }else xx = (int) (t*t);
    } 
    if (y[0]<0) {
      printf("Error: (gauged) y[0] is negative. Increasing t0 may solve this trouble. Break.\n"); 
      printf("t=%.15e t^2=%.15e, gauged y[0]=%.15e, y[1]=%.15e, y[2]=%.15e, bare y[0]=%.15e,Hkn_c0=%.15e\n", t,t*t, y[0]*Hkn_c0, y[1]*Hkn_c1,y[2]*Hkn_c2,y[0],Hkn_c0); 
      break;
    }
  }

  gsl_odeiv_evolve_free(e);
  gsl_odeiv_control_free(c);
  gsl_odeiv_step_free(s);
  if (Hkn_fname!=NULL) {output_value(t,y,t0,h);output_value2(t,y,t0,h);}
  if (Hkn_file != NULL) output_im_value(-1.0,NULL);
  return 0;
}
 

gsl_odeiv_control *setparam(int argc,char *argv[],
                            double *tp,double *t1p,double *hp,double y[],int *show_diffp) {
  gsl_odeiv_control *c=NULL;
  int i,j;
  double abs_err,rel_err;
  double t,tt;
  char tmp_file[1024*10];
  abs_err=Hkn_abs_err;
  rel_err=Hkn_rel_err;
  Hkn_k=3;
  Hkn_n=1;
  Hkn_z=10000;
  *tp=2000; *t1p=2050; *hp=1e-5;
  y[0]=1.0092527990727128287e15;
  y[0] = y[0]/exp(2*(Hkn_k+1)*log(*tp)); // a27c-x.c --> 27d-x.c Gauge 変更.
  y[1]=0.000063078313247623732261;
  y[2]=-0.0000315391568209;

  for (i=1; i<argc; i++) {
    if (strcmp(argv[i],"--help")==0) {
      printf("a27c-x [--t0 start_t --t t --init iv0 iv1 iv2 --abs_err num1 --rel_err num2 --k k --n n --y y --fname name1 --verbose --quiet --show_diff --gauge2] [--file datafile --pos_i i --pos_j j --file_max max_size --file_from from_x --file_step step_x --logC log_of_scalar_gauge_factor\n");
      printf(" returns H^k_n(t^2,y) from the value at t=start_t\n");
      printf("Initial value --init must be shifed by the gauge exp(-t^2+2zt+2(k+1)log(t))\n");
      printf("If --gauge2 init is shifted by the gauge exp(-t^2*2zt+2(k+1)*log(t)) where t = maxarg of arg of exp. This option must be after the --y option.\n");
      printf("Let C be log_of_scalar_gauge_factor. Input initial *vector* should be multiplied by exp(-C). Output value y[0] will be multiplied by exp(C)\n");
      return(c);
    }
    if (strcmp(argv[i],"--t0")==0) {
       i++; sscanf(argv[i],"%lg",tp);
    }else if (strcmp(argv[i],"--t")==0) {
       i++; sscanf(argv[i],"%lg",t1p);
    }else if (strcmp(argv[i],"--h")==0) {
       i++; sscanf(argv[i],"%lg",hp);
    }else if (strcmp(argv[i],"--init")==0) {
       for (j=0; j<3; j++) sscanf(argv[++i],"%lg",&(y[j]));
    }else if (strcmp(argv[i],"--abs_err")==0) {
       i++; sscanf(argv[i],"%lg",&abs_err);
    }else if (strcmp(argv[i],"--rel_err")==0) {
       i++; sscanf(argv[i],"%lg",&rel_err);
    }else if (strcmp(argv[i],"--k")==0) {
       i++; sscanf(argv[i],"%lg",&Hkn_k);
    }else if (strcmp(argv[i],"--n")==0) {
       i++; sscanf(argv[i],"%lg",&Hkn_n);
    }else if (strcmp(argv[i],"--y")==0) {
       i++; sscanf(argv[i],"%lg",&Hkn_z); Hkn_z = sqrt(Hkn_z);
    }else if (strcmp(argv[i],"--logC")==0) {
       i++; sscanf(argv[i],"%lg",&Hkn_logC); 
    }else if (strcmp(argv[i],"--fname")==0) {
       i++; Hkn_fname = (char *)malloc(strlen(argv[i])+2);
       strcpy(Hkn_fname,argv[i]);
    }else if (strcmp(argv[i],"--show_diff")==0) {
      if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
        i++; sscanf(argv[i],"%d",show_diffp);
      }else{
        *show_diffp=1;
      }
    }else if (strcmp(argv[i],"--file")==0) {
       i++; Hkn_file = (char *)malloc(strlen(argv[i])+1024);
       strcpy(tmp_file,argv[i]);
    }else if (strcmp(argv[i],"--pos_i")==0) {
       i++; sscanf(argv[i],"%d",&Hkn_i);
    }else if (strcmp(argv[i],"--pos_j")==0) {
       i++; sscanf(argv[i],"%d",&Hkn_j);
    }else if (strcmp(argv[i],"--file_from")==0) {
       i++; sscanf(argv[i],"%d",&Hkn_file_from);
    }else if (strcmp(argv[i],"--file_step")==0) {
       i++; sscanf(argv[i],"%d",&Hkn_file_step);
    }else if (strcmp(argv[i],"--file_max")==0) {
       i++; sscanf(argv[i],"%d",&Hkn_file_max);

    }else if (strcmp(argv[i],"--verbose")==0) {
      Verbose=1;
    }else if (strcmp(argv[i],"--quiet")==0) {
      Verbose=0;
    }else if (strcmp(argv[i],"--gauge2")==0) {
      Hkn_init_gauge=2;
      Hkn_switch_gauge=2;  /*  (***), see also a27d-init.rr */
      // t = (Hkn_z/2.0)*(1+sqrt(1+4*(Hkn_k+1)/(Hkn_z*Hkn_z))); 
      t = *tp;
      tt=-t*t+2*t*Hkn_z+2*(Hkn_k+1)*log(t);
      Hkn_logc0 = tt; // --gauge2 option must come after --y, --k, --t0
    }else {
      fprintf(stderr,"Error: Unknown option %s\n",argv[i]);
      return(NULL);
    }
  }
  Hkn_abs_err = abs_err;
  Hkn_rel_err = rel_err;
  if (c == NULL) c = gsl_odeiv_control_y_new(abs_err, rel_err);
  if (Hkn_file != NULL) {
    sprintf(Hkn_file,"tmp-Hkn-%s-%d-%d.rr",tmp_file,Hkn_i,Hkn_j);
  }
  return(c);
}

int output_value(double t,double y[],double t0,double h) {
  FILE *fp;
  int i;
  double cc[3];
  cc[0] = Hkn_c0;
  cc[1] = Hkn_c1; cc[2] = Hkn_c2;
  if (strcmp(Hkn_fname,"stdout")==0) fp = stdout;
  else fp = fopen(Hkn_fname,"w");
  if (fp == NULL) {fprintf(stderr,"fail to open\n"); return(-2);}
  fprintf(fp,"K=%d$ N=%d$ T=%lg$ Y=%lg$ T_start=%lg$ H=%lg$ Err=[%lg,%lg]$ Hkn_z=%lg$ Hkn_logc0=%lg$ Gauged_y0=%lg$\n",
              (int)Hkn_k,(int)Hkn_n,t,Hkn_z*Hkn_z,t0,h, Hkn_abs_err,Hkn_rel_err,
          Hkn_z,Hkn_logc0,cc[0]*y[0]);
  fprintf(fp,"T_val=[");
  switch(Hkn_switch_gauge) {
  case 1:  // use exp(-t^2+2zt+2(k+1)*log(t))
#define OUT_Hkn  fprintf(fp,"exp(-(%.15e)*(%.15e)+2*(%.15e)*(%.15e)+2*(%.15e+1)*(%.15e))*(%.15e)*exp(%.15e),",t,t,t,Hkn_z,Hkn_k,log(t),y[0]*cc[0],Hkn_logC) 
    OUT_Hkn ;
    break;
  case 2:  // use exp(Hkn_log_c0)
#define OUT_Hkn_case2 fprintf(fp,"exp(%.15e)*(%.15e)*exp(%.15e),",Hkn_logc0,y[0]*cc[0],Hkn_logC);
    OUT_Hkn_case2 ;
    break;
  default:
    fprintf(stderr,"Error, unknown Hkn_switch_gauge mode.\n");
    break;
  }
#define OUT_other  fprintf(fp,"(%.15e)*exp(%.15e)",y[i]*cc[i],Hkn_logC)
  for (i=1; i<3; i++) {
    OUT_other ;
    if (i != 2) fprintf(fp,", ");
    else fprintf(fp,"]$\nend$\n");
  }
  if (fp != stdout) fclose(fp);
  return(0);
}
int output_value2(double t,double y[],double t0,double h) {
  FILE *fp;
  int i;
  double cc[3];
  char fname[1024];
  strcpy(fname,Hkn_fname); strcat(fname,"r");

  cc[0] = Hkn_c0;
  cc[1] = Hkn_c1; cc[2] = Hkn_c2;
  if (strcmp(Hkn_fname,"stdout")==0) fp = stdout;
  else fp = fopen(fname,"w");
  if (fp == NULL) {fprintf(stderr,"fail to open\n"); return(-2);}
  fprintf(fp,"[");
  switch(Hkn_switch_gauge) {
  case 1:  // use exp(-t^2+2zt+2(k+1)*log(t))
    OUT_Hkn ;
    break;
  case 2:  // use exp(Hkn_log_c0)
    OUT_Hkn_case2 ;
    break;
  default:
    fprintf(stderr,"Error, unknown Hkn_switch_gauge mode.\n");
    break;
  }
  for (i=1; i<3; i++) {
    OUT_other ; 
    if (i != 2) fprintf(fp,", ");
    else fprintf(fp,"]\n");
  }
  if (fp != stdout) fclose(fp);
  return(0);
}

// 以下は答え合わせ用.
int showdiff(int x,double t, double y[]) {
  // 答え合わせ用  a30-test4-out.txt
  double y2010[4] = {1.0438126889111481018e15,0.000062921206369448202107,-0.0000314606033804};
  double y2020[4] = {1.0793840305720544204e15,0.000062765267575159058491,
                     -0.0000313826339818};
  double y8000[4] = {3.3071191766976384416e19,0.000031539156549891966725,
                     -0.0000157695783};
  double y9000[4] = {1.4222338329057638033e20,0.000029735401956529103318,
                     -0.000014867700999};
  double y10001[4] = {9.212830058723463133e23,0.000028208068826833957408,
                      -0.000014104034431};
  double y10002[4] = {9.976578620660536556e23,0.000028206658670167388448,
                      -0.000014103329353};
  double y11000[4] = {9.999998705185010936e23,0.000026896683073024013355,
                      -0.000013448341552};

  if (x == 2010) { show_diff2_case1(x,t,y,y2010); }
  if (x == 2020) { show_diff2_case1(x,t,y,y2020); }
  if (x == 8000) { show_diff2_case1(x,t,y,y8000); }
  if (x == 9000) { show_diff2_case1(x,t,y,y9000); }
  if (x == 10001) { show_diff2_case1(x,t,y,y10001); }
  if (x == 10002) { show_diff2_case1(x,t,y,y10002); }
  if (x == 11000) { printf("Reference value is not trustful. "); show_diff2_case1(x,t,y,y11000); }
  
  return(0);
}
/* it works for the case2, too */
int show_diff2_case1(int x, double t, double y[],double y_ans[]) {
  y_ans[0] = y_ans[0]/exp(2*(Hkn_k+1)*log(t));
  double ratio=(y[0]*Hkn_c0-y_ans[0])/y_ans[0];
  if (ratio < 0) ratio = -ratio;
  printf("t=%.15e, (gauged) y[0]=%.15e, y[1]=%.15e, diff y(%d)[0]=%.15e, err/ans=%.15e", t, y[0]*Hkn_c0,
         y[1]*Hkn_c1,x,y[0]*Hkn_c0-y_ans[0],ratio);
  if (ratio > 1e-3) printf(", error is more than 1e-3\n");
  else printf("\n");
  return(0);
}

int output_im_value(double t,double y[]) {
  static FILE *fp = NULL;
  double cc[3];
  int i;
  int xx;
  double x;
  static int size=0;
  static int count=0;
  static double y0_prev=0;
  int rapid_change=0;
  if (y == NULL) { 
    if (fp != NULL) {
      fprintf(fp,"0]$\n");
      fclose(fp);
    }
    return(0);
  }
  if (count % Hkn_file_step == 0) {
    rapid_change=0;
  }else if ((y0_prev != 0) && ((fabs(y[0]/y0_prev) > RAPID_CHANGE) ||
            (fabs(y[0]/y0_prev) < RAPID_CHANGE2)) ) {
    rapid_change=1;
  }
/*  Test for rapid.  See 2018.08.11
./a27d-x --t 2100 --show_diff 2 --file hoge --pos_i 1 --pos_j 2 --file_step 30000
with RAPID_CHANGE 1 pm 1e-10
*/
//  if(rapid_change) printf("change=%lg, y[0]=%lg, y0_prev=%lg\n",fabs(y[0]/y0_prev),y[0],y0_prev);
  if (y0_prev==0) y0_prev=y[0];
  if (rapid_change) y0_prev=y[0];
  x=t*t; xx = (int) x;
  if (((count % Hkn_file_step) != 0) &&(t>=0)) {
    if (!rapid_change) {count++; return(0);}
  } else {count=0; }
  if ((size > Hkn_file_max)&&(t>=0)) return(-2); else {
   if (!rapid_change) size++;
  }
  if (fp == NULL) {
    if ((fp = fopen(Hkn_file,"w"))==NULL) {
      fprintf(stderr,"Open error of %s\n",Hkn_file);
      return(-1);
    }
    fprintf(fp,"H_%d_%d=[\n",Hkn_i,Hkn_j);
  }
  if ((t<0) && (fp != NULL)) {
    fprintf(fp,"0]$\n");
    fclose(fp);
    return(0);
  }

  cc[0] = Hkn_c0; cc[1] = Hkn_c1; cc[2] = Hkn_c2;
  fprintf(fp,"[%d,%.15e,%.15e,",xx, t, x);
  switch(Hkn_switch_gauge) {
  case 1:  // use exp(-t^2+2zt+2(k+1)*log(t))
    OUT_Hkn ;
    break;
  case 2:  // use exp(Hkn_log_c0)
    OUT_Hkn_case2 ;
    break;
  default:
    fprintf(stderr,"Error, unknown Hkn_switch_gauge mode.\n");
    break;
  }
  for (i=1; i<3; i++) {
    OUT_other ;
    if (i != 2) fprintf(fp,", ");
    else fprintf(fp,"],\n");
  }
  if (!rapid_change) count++;
  return(0);
}

// ./a27c-x --t 10010 --show_diff 2 
// ./a27c-x --t0 8000 --t 11010 --init 3.3071191766976384416e19 0.000031539156549891966725 -0.0000157695783 --show_diff 2
// ./a27c-x --t 2100 --show_diff 2 --file hoge --pos_i 1 --pos_j 2

